

'// FOXPRO specific memo handling routines                             

'// Normally, FoxPro memos are somewhat inefficent because existing memos are
'// left in the memo file even after they are deleted. Cheetah uses a clever
'// deleted memo garbage chain that tracks these deleted memos in order that
'// the occupied space can be re-used when a new memo is added. The garbage
'// chain is stored in the unused 495 byte portion of the memo header.


'//
'//  Generic routine to set the internal memo buffer for a
'//  specified field. The data is set depending on whether
'//  it is supplied by the user (xdbSetField) or from the
'//  memo file itself (xdbGetField)  
'//
Function ctx_SetMemoBuffer_FoxPro_Internal( ByVal xdb As CHEETAH4_TYPE Ptr, _
                                            ByVal nSetMode As Long _
                                            ) As Long 
                                            
                                            
   Local pDBF         As DBF_TYPE Ptr 
   Local pField       As DBF_FIELD_TYPE Ptr 
   Local pBlockHeader As DBF_MEMOBLOCK_TYPE Ptr
   Local nFieldNum    As Long                              
   Local sTempBlock   As String * 4096  ' enough for 8, %MEMO_BLOCK_SIZE byte blocks
   Local sEntireMemo  As String
   Local nRecOffset   As Long
   Local nNumBytes    As Long
   Local nBlockSize   As Long
   Local nBlocksRead  As Long
   Local nMemoLength  As Long
   
   pDBF = @xdb.reserved

   nFieldNum = @xdb.nData4byte - 1  ' b/c pointer array is zero based
   
   pField = @pDBF.pFields
   
   nBlockSize = SwapBytesWord( @pDBF.DBH_MEMO.BlockSize )
   
   If @pField[nFieldNum].OrigFirstBlock Then
      ' Retrieve all the blocks for this memo. The first block will contain
      ' the memo header that will indicate the length of the memo.
      nRecOffset = (@pField[nFieldNum].OrigFirstBlock * nBlockSize) + 1
      nNumBytes  = SizeOf(sTempBlock)
      If fGetAt( @pDBF.hFileMemo, nRecOffset, nNumBytes, VarPtr(sTempBlock) ) Then
         @xdb.errcode = %ERR_FILEACCESS: GoTo ExitOut
      End If 
      
      
      pBlockHeader = VarPtr(sTempBlock)
      nMemoLength  = SwapBytesDWord( @pBlockHeader.MemoLength )
      
      @pField[nFieldNum].OrigNumBlocks = Ceil( (nMemoLength + SizeOf(DBF_MEMOBLOCK_TYPE) ) / nBlockSize )
      
      nBlocksRead = ( SizeOf(sTempBlock) / nBlockSize ) 
      
      sEntireMemo = sTempBlock   ' save what we have read so far

      ' If we did not read enough blocks the first time then read them now.
      Do Until nBlocksRead > @pField[nFieldNum].OrigNumBlocks 
         nRecOffset = (@pField[nFieldNum].OrigFirstBlock * nBlockSize) + SizeOf(sTempBlock) + 1
         nNumBytes  = SizeOf(sTempBlock)
         If fGetAt( @pDBF.hFileMemo, nRecOffset, nNumBytes, VarPtr(sTempBlock) ) Then
            @xdb.errcode = %ERR_FILEACCESS: GoTo ExitOut
         End If           
         sEntireMemo = sEntireMemo + sTempBlock
         nBlocksRead = nBlocksRead + ( SizeOf(sTempBlock) / nBlockSize )  
      Loop
      
      sEntireMemo = Left$( sEntireMemo, @pField[nFieldNum].OrigNumBlocks * nBlockSize )
      
   End If

   
   Select Case As Long nSetMode             
      
      Case %CTX_MEMO_USER
         ' Allocate memory in the Field structure to hold the memo.
         nMemoLength = Ceil( (@xdb.nLen + SizeOf(DBF_MEMOBLOCK_TYPE)) / nBlockSize ) * nBlockSize
         @pField[nFieldNum].pMemoText = MemAlloc( nMemoLength )                          
         
         ' Set the block header values
         pBlockHeader = @pField[nFieldNum].pMemoText
         @pBlockHeader.Signature  = SwapBytesDWord( 1 )
         @pBlockHeader.MemoLength = SwapBytesDWord( @xdb.nLen )
         
         ' Copy the user data into the memo buffer
         MemCopyD @xdb.nData, _
                  @pField[nFieldNum].pMemoText + SizeOf(DBF_MEMOBLOCK_TYPE), _
                  @xdb.nLen                         
                  
         ' Set the parameters for this memo. 
         @pField[nFieldNum].MemoTextLength = SizeOf(DBF_MEMOBLOCK_TYPE) + @xdb.nLen 
         @pField[nFieldNum].MemoDirty      = %TRUE
         

      Case %CTX_MEMO_DISK
         @pField[nFieldNum].pMemoText      = MemAlloc( Len(sEntireMemo) )  
         MemCopyD StrPtr(sEntireMemo), _
                  @pField[nFieldNum].pMemoText, _
                  Len(sEntireMemo)
                                 
         @pField[nFieldNum].MemoTextLength = SizeOf(DBF_MEMOBLOCK_TYPE) + nMemoLength
         @pField[nFieldNum].MemoDirty      = %FALSE   ' true only when user changes it

   End Select   
   

ExitOut:
   Function = @xdb.errcode 

End Function



'//
'//  Generic routine to write the internal memo fields to disk. 
'//
Function ctx_WriteMemoBuffer_FoxPro_Internal( ByVal xdb As CHEETAH4_TYPE Ptr ) As Long
                                            
   Local pDBF        As DBF_TYPE Ptr
   Local pField      As DBF_FIELD_TYPE Ptr 
   Local pChain      As DBF_GARBAGEMEMO_TYPE Ptr
   Local x           As Long
   Local y           As Long  
   Local fUsingChain As Long
   Local nFoundat    As Long
   Local nRecOffset  As Long
   Local nNumBytes   As Long
   Local nNumBlocks  As Long                 
   Local nStartBlock As Long
   Local nBlockSize  As Long  
   
   ' Loop through the fields and output the dirty memo buffers to disk. The
   ' RecordBuffer is updated with the correct starting block number after
   ' the write is finished. This is the reason that the memo fields need to
   ' be written prior to the RecordBuffer being saved into the database.

   pDBF   = @xdb.reserved
   pField = @pDBF.pFields

   nBlockSize = SwapBytesWord( @pDBF.DBH_MEMO.BlockSize )
      
   For x = 1 To @pDBF.NumFields
          
      If ctx_IsMemoField_Internal( pField ) Then
         If @pField.MemoDirty = %TRUE Then 
            GoSub InvalidateBlocks
            GoSub WriteBlocks
         End If
      End If
      
      Incr pField
   Next
   
      
ExitOut:
   Function = @xdb.errcode 

Exit Function
   
   
   
InvalidateBlocks:
   ' Invalidate any existing blocks for this memo. This simply entails
   ' updating the memo garbage chain. The new blocks are written to
   ' the end of the memo file. (The garbage chain is unique to Cheetah
   ' in that it uses wasted space in the dbase memo header for storage)
   
   ' If no original number of blocks existed for this memo then it was
   ' never previously saved to disk (i.e. it is a new record, or an existing
   ' record without any previously saved memo text).
   If @pField.OrigFirstBlock = 0 Then Return  ' nothing to invalidate
   
   ' If the number of blocks has not yet been established then retrieve them now.
   If @pField.OrigNumBlocks = 0 Then
      If ctx_SetMemoBuffer_foxpro_Internal( xdb, %CTX_MEMO_DISK ) Then GoTo ExitOut
   End If   
   
   ' Determine where to place the deleted blocks. If there is no
   ' more room then do nothing. If the blocks are adjacent to an
   ' existing garbage block then simply add it to that existing
   ' chain. 
   nFoundat = 0
   For y = 1 To %MAX_GARBAGECHAIN
      ' In case we can not add the block to an existing chain then
      ' we need to track into which free position to put the information.
      If ( @pDBF.DBH_MEMO.GarbageChain(y).BlockNumber = 0 ) And ( nFoundat = 0 ) Then
         nFoundat = y: Iterate For
      End If   

      ' Attempt to add the blocks to an existing chain
      If @pDBF.DBH_MEMO.GarbageChain(y).BlockNumber + _
         @pDBF.DBH_MEMO.GarbageChain(y).NumGarbageBlocks = @pField.OrigFirstBlock Then
      
         @pDBF.DBH_MEMO.GarbageChain(y).NumGarbageBlocks =  _
            @pDBF.DBH_MEMO.GarbageChain(y).NumGarbageBlocks + @pField.OrigNumBlocks
         nFoundat = 0: Exit For
      End If   
   Next
   If nFoundat > 0 Then
      @pDBF.DBH_MEMO.GarbageChain(nFoundat).BlockNumber      = @pField.OrigFirstBlock
      @pDBF.DBH_MEMO.GarbageChain(nFoundat).NumGarbageBlocks = @pField.OrigNumBlocks
   End If
   
Return


WriteBlocks:
   ' Write the new data to the memo file. Use any existing deleted blocks
   ' in the garbage chain. Therefore, if the same number of blocks are
   ' needed in the new write that existed in the original data, then we
   ' are simply writing to the same blocks. If the new data is larger then
   ' the memo must be written to the end of the file.
  
   ' Determine if we are deleting an existing memo. In that case, the memo
   ' length will be zero (after the length of the block header is removed). 
   If (@pField.MemoTextLength - SizeOf(DBF_MEMOBLOCK_TYPE)) <= 0 Then
      nStartBlock = 0
      nNumBlocks  = 0  
      
   Else   
   
      ' Get the number of blocks to write.
      nNumBlocks = Ceil( @pField.MemoTextLength / nBlockSize )
      
      nStartBlock = SwapBytesDWord( @pDBF.DBH_MEMO.NextFreeBlock )  ' default
   
      ' Determine the block to start writing at.
      ' Try to use a deleted chain that has the correct amount of free blocks. If
      ' the exact amount does not exist then use the one that has at least the
      ' same number of blocks as needs to be written.
      fUsingChain = %FALSE
      nFoundat = 0
      pChain   = VarPtr(@pDBF.DBH_MEMO.GarbageChain(1))
      For y = 1 To %MAX_GARBAGECHAIN
         If @pChain.NumGarbageBlocks = nNumBlocks Then
            nFoundat    = 0
            nStartBlock = @pChain.BlockNumber
            fUsingChain = %TRUE
            ' reset this garbage chain entry so that it can be re-used in the future
            @pChain.NumGarbageBlocks = 0
            @pChain.BlockNumber      = 0
            Exit For
         ElseIf @pChain.NumGarbageBlocks >= nNumBlocks Then
            nFoundat = y
         End If   
         Incr pChain
      Next  
      
      ' Adjust a found garbage chain that contains more blocks
      ' then we need to write.
      If nFoundat > 0 Then
         fUsingChain = %TRUE
         nStartBlock = @pDBF.DBH_MEMO.GarbageChain(nFoundat).BlockNumber 
         @pDBF.DBH_MEMO.GarbageChain(nFoundat).BlockNumber = _
               @pDBF.DBH_MEMO.GarbageChain(nFoundat).BlockNumber + nNumBlocks
         @pDBF.DBH_MEMO.GarbageChain(nFoundat).NumGarbageBlocks =  _
               @pDBF.DBH_MEMO.GarbageChain(nFoundat).NumGarbageBlocks - nNumBlocks
      End If
      
      nRecOffset = (nStartBlock * nBlockSize) + 1
      nNumBytes  = nNumBlocks * nBlockSize

      If fWriteAt( @pDBF.hFileMemo, nRecOffset, nNumBytes, @pField.pMemoText ) Then
         @xdb.errcode = %ERR_FILEACCESS: GoTo ExitOut
      End If
   
      If fUsingChain = %FALSE Then
         @pDBF.DBH_MEMO.NextFreeBlock = SwapBytesDWord( nStartBlock + nNumBlocks )                                                               
      End If
   
   End If
   
   ' Write the memo header to disk.
   If ctx_WriteMemoHeader_Internal( xdb ) Then GoTo ExitOut
   
   ' Update the RecordBuffer with the starting block number.
   Poke Long, @pDBF.pRecordBuffer + @pField.FDA, nStartBlock
   
   ' Update the Field structure with the new saved memo data.
   @pField.OrigFirstBlock = nStartBlock
   @pField.OrigNumBlocks  = nNumBlocks
                                                                  
Return

End Function


